// XIO.c
#include <propeller.h>
#include <STDIO.H>
#include <STRING.H>
#include <STDLIB.H>
#include "BufferedSerial.h"
#include "i2c.h"


// 31 / 30 is the standard for RxD / TxD 
#define RS232RxD 31 /* my setting: P25, input */
#define RS232TxD 30 /*             P27, output */
#define BAUDRATE 115200  /* also tested with the primitive RS232 */

#define EEPROMPageSize 64  /* setting this bigger doesn't improve things at 115200 baud. Making it smaller only takes more
                            processing time. */

typedef enum {normal, autoSPIN, nvVar} eLengthMode_t; // BUGBUG: nvVar not jet implemented

static int verbose = FALSE;
static int leadingAddress = FALSE;

int devAddr = 28 << 18;  // the Propeller-default boot EEPROM 
int pingPong = FALSE;
int i2cRunning = FALSE;
long offset = 0;

struct {
  int wNdx;
  int entries;
  long addr;
  char buffer[EEPROMPageSize];
} writeBuffer;

// -------------------------------------
// for printf
int putchar(char ch) {
  SerialTx(ch);
  return ch;
}
// -------------------------------------
// for gets
int getchar(void) {
  return SerialRx();
}
// ----------------------------------
void Help(void) {
  printf("\n\n3PClient   3-decker propeller client for reading / writing to EEPROM (only i2c)\n");
  printf("(c) 2009 by Nick Mueller / Puchschlagen / Germany\n\n");
  printf("Lines must ONLY end with an Lf (0x0A)\n");
  printf("Values in <> can be decimal, hex (0x...) or binary (0b...)\n\n");
  printf("Commands:\n");
  printf("-? : This screen\n");
  printf("// : comment, skipped only during receiving commands. Must be at the very left\n");
  printf("-v : verbose mode (for debugging, default = off)\n");
  printf("--v: verbose off\n");
  printf("-p : ping-pong mode on. ping-pong mode sends a '-r' back as an '-w'\n");
  printf("     for later playback\n");
  printf("--p: ping-pong mode off (default).\n\n");
  printf("-a : Add leading address-field at each line (for '-r').\n");
  printf("    '-w'-command ignores that field. Default is off\n");
  printf("--a: Switch off '-a' command\n"); 
  printf("-d:<Device> : device (EEPROM) address.\n    defaults to 28 << 18 (the Prop boot EEPROM)\n");
  printf("-r:<Addr>:<Length> : Read EEPROM at <Address>, <Length> bytes\n");
  printf("   if <Length> = 'a' (auto), it is assumed that <Address> points to an\n");
  printf("   executable image and the length is taken from that image.\n");
  printf("-w:<Addr>:<Length> : Write EEPROM (rest like '-r'\n");
  printf("-o:<Offset> : Mainly for playback, adds offset <Offset> to address\n");
  printf("              reading (-r) Adds offset to the dumped -NOT the reading- address\n");
  printf("              writing (-w) Adds an offset to the writing address\n\n");
}
// ----------------------------------
// converts string (in d, h, b) to int, returns pointer to end
char* GetValue(char* pStr, long* pValue) {
  long tmp;
  char* pEnd = pStr; 
  
  tmp = strtol(pStr, &pEnd, 0);
  if ((tmp != LONG_MAX) & (tmp != LONG_MIN))
    *pValue = tmp;
  
  return pEnd;
}
// ----------------------------------
unsigned int GetLength(long addr, eLengthMode_t mode) {
  unsigned short length;
  
  if (mode == autoSPIN) {
    i2cReadEEPROM(addr + devAddr + 14, &length, sizeof(length));
    return length - 4;
  }  
  
  return 0;
}
// ----------------------------------
void ReadEEPROM(long addr, long length, eLengthMode_t mode) {
  int i;
  unsigned int byte = 0; // must be int for printf
  unsigned int check;
  

  // BUGBUG, missing code in GetLength
  // <Length> = 'n' auto-size for NVVar
  addr+= devAddr;
  
  if (mode == nvVar)
    printf("// Image of NVVar-area (C format V0.1). Length : %d\n", length);
  if (mode == autoSPIN)
    printf("// Image of executable code. Length : %d\n", length);

  for (i = 0; i < length; i++) {
    if (i % 16 == 0) { // new line started
      check = 0;
      if (leadingAddress)
        printf("0x%05X:", addr - devAddr + offset);
    }
    i2cReadEEPROM(addr++, &byte, 1);
    printf("%02X", byte);
    check += byte;
    if (i % 16 == 15)
      printf(" %02X\n", check % 0xFF);
  }
  
  // print end of line, perhaps
  if (length % 16 != 0)
    printf(" %02X\n", check % 0xFF);
}
// ----------------------------------
void StartBuffer(long addr) {
  writeBuffer.wNdx = 0;
  writeBuffer.addr = addr;
  writeBuffer.entries = 0;
}
// ----------------------------------
int FlushBuffer(void) {
  int i;
  char checkBuffer[EEPROMPageSize];
  int pageRest = EEPROMPageSize - writeBuffer.addr % EEPROMPageSize;

  i2cWriteWait(devAddr);
  i2cWriteEEPROM(writeBuffer.addr, &(writeBuffer.buffer[0]), writeBuffer.entries);
  i2cWriteWait(devAddr);
  i2cReadEEPROM(writeBuffer.addr, &checkBuffer[0], writeBuffer.entries);
  i2cWriteWait(devAddr);
    
  for (i = 0; i < writeBuffer.entries; i++) {
    if (writeBuffer.buffer[i] != checkBuffer[i])
      return FALSE;
  }
  SerialTx('+');

  writeBuffer.addr+= writeBuffer.entries;
  writeBuffer.wNdx = 0;
  writeBuffer.entries = 0;

  return TRUE;
}
// ----------------------------------
int WriteBuffer(char byte) {
  int pageRest = EEPROMPageSize - writeBuffer.addr % EEPROMPageSize;
  
  writeBuffer.buffer[writeBuffer.wNdx++] = byte;
  writeBuffer.entries++;

  if (writeBuffer.entries >= pageRest) // either the part that fits into the page rest, or a complete page
    return FlushBuffer();
  
  return TRUE;
}
// ----------------------------------
int DeNibble(char ch) {
  if ((ch >= '0') & (ch <= '9'))
    return ch - '0';
    
  if ((ch >= 'A') & (ch <= 'F'))
    return ch - 'A' + 10;
    
  if ((ch >= 'a') & (ch <= 'f'))
    return ch - 'a' + 10;
    
  return 0x100; // force an overflow
}
// ----------------------------------
void WriteEEPROM(long addr, long length) {
  int i;
  unsigned int byte = 0; // must be int for printf
  unsigned int check;
  char inBuffer[200];
  char* pChar;
  int stopNdx;
  
  addr+= devAddr + offset;
  
  StartBuffer(addr);
  
  // the address-field at the beginning of the line is ignored, because all information
  // is already given in addr & length
  
  stopNdx = length - 1;
  
  i = 0;
  for (;;) { // all lines
    do { // skip comments
      pChar = gets(inBuffer);
    } while (strncmp("//", inBuffer, 2) == 0);
   
    // ignore the address-field if it exists
    pChar = strchr(inBuffer, ':');
    if (pChar != NULL)
      pChar++;
    else
      pChar = &inBuffer[0];

    check = 0;    
    for (;;) {  // do one complete line
      byte = (DeNibble(*(pChar++)) << 4) + DeNibble(*(pChar++));

      if (byte > 0xFF) { // got an error
        printf("\nabort\n// invalid value (non-HEX) near:'%s'\n", inBuffer); //pChar - 2);
        return;
      } 
        
      check += byte;

      if (!WriteBuffer(byte)) {
        printf("\nabort\n");
        printf("// Error verifying during write near:0x%x\n", inBuffer);
        return;
      }

      i++;
      if ((i % 16 == 0) | (i > stopNdx)) { // end of line, or end of transmition. Write 16 bytes to EEPROM
        unsigned int lCheck;
        
        // read the checksum
        while (*pChar == ' ') // skip any leading blanks in front of the checksum
          pChar++;
          
        lCheck = (DeNibble(*(pChar++)) << 4) + DeNibble(*pChar); // no need for second ++, pChar isn't used anymore

        if (lCheck != check % 0xFF) {
          printf("\nabort\n// checksum-error 0x%2X : 0x%2X in '%s'\n", check, lCheck, inBuffer);
          return;
        }
        
        break;
      }  // one complete line
    }
    
    if (i > stopNdx)
      break;    
  }
  if (!FlushBuffer()) {
    printf("\nabort\n// Error verifying during write near:0x%x\n", inBuffer);
    return;
  }
    
  printf("\n// done writing EEPROM, no errors\n");
}
// ----------------------------------
void ReceiveCommand(void) {
  #define DecodeBufferSize 100
  char decodeBuffer[DecodeBufferSize];
  long addr;
  long length;
  char* ptr;
  eLengthMode_t mode;
  
  // Each line terminates with an 0x0D
  // Numbers can be decimal, hex (0x...) or binary (0b...)
  
  if (offset != 0)
    printf("// *** Warning: Offset (%d, 0x%x) active\n", offset, offset);

  for (;;) {
    gets(&decodeBuffer[0]);
      
    // verbose
    if (strcmp("-v", decodeBuffer) == 0) {
      verbose = TRUE;
      return;
    } 

    if (strcmp("--v", decodeBuffer) == 0) {
      verbose = FALSE;
      return;
    } 
      
    // ping-pong on
    if (strcmp("-p", decodeBuffer) == 0) {
      pingPong = TRUE;
      return;
    }
      
    // ping-pong off
    if (strcmp("--p", decodeBuffer) == 0) {
      pingPong = FALSE;
      return;
    }
      
    // Help screen
    if (strcmp("-?", decodeBuffer) == 0) {
      Help();
      return;
    }
     
    // address-field
    if (strcmp("-a", decodeBuffer) == 0) {
      leadingAddress = TRUE;
      return;
    }
    if (strcmp("--a", decodeBuffer) == 0) {
      leadingAddress = FALSE;
      return;
    }
      
    // offset
    if (strncmp("-o:", decodeBuffer, 3) == 0) {
      GetValue(&decodeBuffer[3], &offset);
      return;
    }
    // skip comment
    if (strncmp("//", decodeBuffer, 2) == 0)
      return;
      
    // read/write command
    if ((strncmp("-r:", decodeBuffer, 3) == 0) | (strncmp("-w:", decodeBuffer, 3) == 0)) {
      char original = decodeBuffer[1];
      mode = normal;
      addr = -1;
      length = 0;
      ptr = GetValue(&decodeBuffer[3], &addr);
      if (strcmp("a", ptr + 1) == 0) {
        length = GetLength(addr, autoSPIN);
        mode = autoSPIN;
        
        // change command for playback
        sprintf(decodeBuffer, "-r:0x%05X:0x%04X", addr, length);
      }
      else
        GetValue(ptr + 1, &length);
      
      if ((addr >= 0) & (length > 0)) {
        if (pingPong) {
          if (decodeBuffer[1] == 'w')
            decodeBuffer[1] = 'r';
          else
            decodeBuffer[1] = 'w';
        }
  
        // echo command for playback with ping-pong adjustment
        printf("%s\n", decodeBuffer);
          
        if (original == 'r')           
          ReadEEPROM(addr, length, mode);
        else
          WriteEEPROM(addr, length);
          
        return;
      }
      if (verbose) {
        if (addr < 0)
          printf("*** invalid address '%X' in '%s'\n", addr, decodeBuffer);
        if (length < 1)
          printf("*** invalid lenght '%d' in '%s'\n", length, decodeBuffer);
      }
    } // read/write
     
    if (verbose) {
      printf("*** Unknown command '%s'. Syncing to next command...\n", decodeBuffer);
      return;
    }    
  }
}
// ----------------------------------
void RS232Setup(void) {
  SerialStart(RS232RxD, RS232TxD, 0, BAUDRATE);
}
// ----------------------------------
void main(void) {
  static i2cControlBlock_t lCtrlBlock;

  i2cStart(&lCtrlBlock);
  RS232Setup();
  wait(10000000);
  SerialRxFlush();

  printf("ready\n");
    
  for (;;)
    ReceiveCommand();  
}

