/**
 * Full Duplex Serial adapter module.
 */
// based on work by Steve Denson (jazzed)
// only minor adjustments made. Mostly for the bigger buffers
// by Nick Mueller
//

#include <propeller.h>
#include <string.h>
#include "BufferedSerial.h"
#include "BufferedSerial_array.h"

static long  gSerialCog = 0;  //cog flag/id

static tSerialStruct gSerial;    // serial descriptor

/* 
 * These buffers must be contiguous. Their size must match the asm expectation.
 * Asm also expects the address of txbuff to be after rxbuff.
 * Apparently the C compiler "allocates" this memory in reverse order.
 * Using a struct would correct it, but for now let's just reverse the entry.
 */
static char  gSerialTxBuff[SERIAL_BUFFER_SIZE];  // transmit buffer
static char  gSerialRxBuff[SERIAL_BUFFER_SIZE];  // receive buffer

/**
 * start initializes and starts native assembly driver in a cog.
 * @param rxpin is pin number for receive input
 * @param txpin is pin number for transmit output
 * @param mode is interface mode. see header FDSERIAL_MODE_...
 * @param baudrate is frequency of bits ... 115200, 57600, etc...
 * @returns non-zero on success
 */

// -------------------------------------------
int SerialStart(int rxpin, int txpin, int mode, int baudrate) {
  return SerialStartFromImage(rxpin, txpin, mode, baudrate, &BufferedSerial_array);
}
// -------------------------------------------
// this StartFromImage is only useful if the code was loaded from EEPROM 
// (and not compiled and read from BufferedSerial_array)
int SerialStartFromImage(int rxpin, int txpin, int mode, int baudrate, void* ptr) {
  // CLKFREQ is internal cpu clock frequency (80000000 for 5M*pll16x)

  SerialStop();
  memset(&gSerial, 0, sizeof(tSerialStruct));

  gSerial.rx_pin = rxpin; // recieve pin
  gSerial.tx_pin = txpin; // transmit pin
  gSerial.mode   = mode;  // interface mode
  gSerial.ticks = CLKFREQ/baudrate; // baud
  gSerial.buffer_size = sizeof(gSerialRxBuff);
  gSerial.rxBufferPtr = &gSerialRxBuff[0];
  gSerial.txBufferPtr = &gSerialTxBuff[0];
  
  gSerialCog = cognew_native(ptr, (void*)&gSerial) + 1;

  wait(10000000); // give cog chance to load
  return gSerialCog;
}
/**
 * stop stops the cog running the native assembly driver 
 */
void SerialStop(void) {
  int id = gSerialCog - 1;
  
  if(gSerialCog > 0) {
    asm("cogstop %id");
  }
}
/**
 * rxflush empties the receive queue 
 */
void SerialRxFlush(void) {
  while(SerialRxCheck() >= 0) {} // clear out queue by receiving all available 
}

/**
 * Gets a byte from the receive queue if available
 * Function does not block. We move rxtail after getting char.
 * @returns receive byte 0 to 0xff or -1 if none available 
 */
int SerialRxCheck(void) {
  int rc = -1;
  
  if(gSerial.rx_tail != gSerial.rx_head) {
    rc = gSerialRxBuff[gSerial.rx_tail];
    gSerial.rx_tail = (gSerial.rx_tail + 1) & (SERIAL_BUFFER_SIZE - 1);
  }
  return rc;
}

/**
 * Get a byte from the receive queue if available within timeout period.
 * Function blocks if no recieve for ms timeout.
 * @param ms is number of milliseconds to wait for a char
 * @returns receive byte 0 to 0xff or -1 if none available 
 */
int SerialRxTime(int ms) {
  int rc;
  int t0;
  int t1;

  asm("mov %t0, cnt");
  do {
    rc = SerialRxCheck();
    asm("mov %t1, cnt");
    if((t1 - t0)/(CLKFREQ/1000) > ms)
      break;
  } while(rc < 0);
  return rc;
}

/**
 * Wait for a byte from the receive queue. blocks until something is ready.
 * @returns received byte 
 */
int SerialRx(void) {
  int rc;

  do
    rc = SerialRxCheck();
  while(rc < 0);
  return rc;
}

/**
 * tx sends a byte on the transmit queue.
 * @param txbyte is byte to send. 
 */
int SerialTx(int txbyte) {
  while(gSerial.tx_tail != gSerial.tx_head) {} // wait for queue to be empty

  gSerialTxBuff[gSerial.tx_head] = txbyte;
  gSerial.tx_head = (gSerial.tx_head+1) & (SERIAL_BUFFER_SIZE - 1);
  if(gSerial.mode & SERIAL_MODE_IGNORE_TX_ECHO)
    return SerialRx(); // why not rxcheck or timeout ... this blocks for char

  return -1;
}

